from pennprov.sql.sql_prov import SqlProvenance
import logging

from pennprov.connection.mprov_connection_cache import MProvConnectionCache
from pennprov.api.decorators import MProvAgg

import pandas as pd

logging.basicConfig(level=logging.DEBUG)
connection_key = MProvConnectionCache.Key()
mprov_conn = MProvConnectionCache.get_connection(connection_key)
if mprov_conn:
    mprov_conn.create_or_reset_graph()
else:
    raise RuntimeError('Could not connect')


def simple_query():
    query = 'select * from test'
    parsed = """
== Parsed Logical Plan ==
'Project [*]
+- 'UnresolvedRelation [test]

== Analyzed Logical Plan ==
x: bigint, y: bigint, name: string
Project [x#0L, y#1L, name#2]
+- SubqueryAlias test
   +- LogicalRDD [x#0L, y#1L, name#2], false

== Optimized Logical Plan ==
LogicalRDD [x#0L, y#1L, name#2], false

== Physical Plan ==
*(1) Scan ExistingRDD[x#0L,y#1L,name#2]
    """
    print(parsed)


def join_query():
    query = 'select t1.x, count(*) from test t1 join test t2 on t1.x = t2.y group by t1.x'
    parsed = """
== Parsed Logical Plan ==
'Aggregate ['t1.x], ['t1.x, unresolvedalias('count(1), None)]
+- 'Join Inner, ('t1.x = 't2.y)
   :- 'SubqueryAlias t1
   :  +- 'UnresolvedRelation [test]
   +- 'SubqueryAlias t2
      +- 'UnresolvedRelation [test]

== Analyzed Logical Plan ==
x: bigint, count(1): bigint
Aggregate [x#0L], [x#0L, count(1) AS count(1)#32L]
+- Join Inner, (x#0L = y#29L)
   :- SubqueryAlias t1
   :  +- SubqueryAlias test
   :     +- LogicalRDD [x#0L, y#1L, name#2], false
   +- SubqueryAlias t2
      +- SubqueryAlias test
         +- LogicalRDD [x#28L, y#29L, name#30], false

== Optimized Logical Plan ==
Aggregate [x#0L], [x#0L, count(1) AS count(1)#32L]
+- Project [x#0L]
   +- Join Inner, (x#0L = y#29L)
      :- Project [x#0L]
      :  +- Filter isnotnull(x#0L)
      :     +- LogicalRDD [x#0L, y#1L, name#2], false
      +- Project [y#29L]
         +- Filter isnotnull(y#29L)
            +- LogicalRDD [x#28L, y#29L, name#30], false

== Physical Plan ==
*(5) HashAggregate(keys=[x#0L], functions=[count(1)], output=[x#0L, count(1)#32L])
+- *(5) HashAggregate(keys=[x#0L], functions=[partial_count(1)], output=[x#0L, count#36L])
   +- *(5) Project [x#0L]
      +- *(5) SortMergeJoin [x#0L], [y#29L], Inner
         :- *(2) Sort [x#0L ASC NULLS FIRST], false, 0
         :  +- Exchange hashpartitioning(x#0L, 200), true, [id=#67]
         :     +- *(1) Project [x#0L]
         :        +- *(1) Filter isnotnull(x#0L)
         :           +- *(1) Scan ExistingRDD[x#0L,y#1L,name#2]
         +- *(4) Sort [y#29L ASC NULLS FIRST], false, 0
            +- Exchange hashpartitioning(y#29L, 200), true, [id=#73]
               +- *(3) Project [y#29L]
                  +- *(3) Filter isnotnull(y#29L)
                     +- *(3) Scan ExistingRDD[x#28L,y#29L,name#30]
    """
    print(parsed)


sql = SqlProvenance(mprov_conn)

print(sql.query_to_entity("test"))

simple_query()
join_query()



import logging
import pytest

from pennprov.connection.mprov import MProvConnection
from pennprov.connection.mprov_connection_cache import MProvConnectionCache
from pennprov.api.decorators import MProvAgg

import pandas as pd

logging.basicConfig(level=logging.DEBUG)
connection_key = MProvConnectionCache.Key()
mprov_conn = MProvConnectionCache.get_connection(connection_key)
if mprov_conn:
    mprov_conn.create_or_reset_graph()
else:
    raise RuntimeError('Could not connect')

sub_stream_1 = mprov_conn.create_collection('output_ecg_1', 1)
sub_stream_2 = mprov_conn.create_collection('output_ecg_2', 1)

@MProvAgg("ecg", 'output_ecg',['x','y'],['x','y'], sub_stream_1)
@pytest.mark.skip(reason="Not a test fn")
def test(n):
    return n.groupby('x').count()

@MProvAgg("ecg", 'output_ecg',['x'],['x'], sub_stream_2)
@pytest.mark.skip(reason="Not a test fn")
def testx(n):
    return n.groupby('x').count()

def test_main():
    # Test the decorators, which will create entities for the dataframe
    # elements, and nodes representing the dataframe components
    ecg = pd.DataFrame([{'x':1, 'y': 2}, {'x':3, 'y':4}])
    test(ecg)
    testx(ecg)

    mprov_conn.store_annotations(sub_stream_1, {'name': 'ecg', 'date': '01-01-01'})
    mprov_conn.store_annotations(sub_stream_2, {'name': 'eeg', 'date': '01-01-05'})

    mprov_conn.flush()

    df = ecg

    id = mprov_conn.get_token_qname(mprov_conn.get_entity_id('output_ecg', 'w[[1, 1], [3, 1]]'))
    node = mprov_conn.get_node(id)
    print ("Output ECG node %s is %s" %(MProvConnection.get_local_part(id), node))

    # What are the sources for a given output ECG (these should be windows)
    source_list = mprov_conn.get_source_entities(mprov_conn.get_token_qname(mprov_conn.get_entity_id('output_ecg', 'w[[1, 1], [3, 1]]')))
    assert(len(source_list) == 1)
    print (str(source_list))
    # Are there any derived?
    derived_list = mprov_conn.get_derived_entities(mprov_conn.get_token_qname(mprov_conn.get_entity_id('output_ecg', 'w[[1, 1], [3, 1]]')))
    assert(len(derived_list) == 0)
    print (str(derived_list))

    # Take a source (window) and see if we can read it
    source_node = mprov_conn.get_node(source_list[0])
    print (str(source_node))
    assert (len(source_node) == 1)# and source_node[0]['type'] == 'QUALIFIED_NAME')
    # and trace to the input tuples
    tuples_list = mprov_conn.get_child_entities(source_list[0])
    assert (len(tuples_list) == 2)
    print (str(tuples_list))

    activity = mprov_conn.get_creating_activities(mprov_conn.get_token_qname(mprov_conn.get_entity_id('output_ecg', 'w[[1, 1], [3, 1]]')))
    print (str(activity))
    assert (len(activity) == 1)
    gen = mprov_conn.get_activity_outputs(activity[0])
    assert (len(gen) == 1)
    print (str(gen))

    input_windows = mprov_conn.get_activity_inputs(activity[0])
    assert (len(input_windows) == 1)
    assert (MProvConnection.get_local_part(input_windows[0]) == "output_ecg_w.w[[1, 1], [3, 1]]")
    print (str(input_windows))

    mprov_conn.store_annotations(gen[0], {'first': 'value', 'second':0})

    mprov_conn.flush()
    ann = mprov_conn.get_annotations(gen[0])
    print ("Annotations on %s: %s" %(str(MProvConnection.get_local_part(gen[0])), str(ann)))
    assert (len(ann) == 2)

    assert (mprov_conn.get_stream_inputs("output_ecg")[0] == "ecg")

    the_producers = mprov_conn.get_stream_producers("output_ecg")

    assert (len(the_producers) == 1)
    assert ('def test' in the_producers[0])

    print (the_producers)

    ####  [i1] \
    ####  [i2] -- [w_{i1,i2,i3}] <-used- (f) <-wasGeneratedBy- [o123]
    ####  [i3] /
    ####


"""
 Copyright 2021 Trustees of the University of Pennsylvania
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from __future__ import print_function

from typing import List, Any, Dict
import hashlib
import binascii
import logging
import re
import pennprov
import datetime
from pennprov.metadata.stream_metadata import BasicTuple

#from pennprov.cache.graph import GraphCache

import psycopg2
from pennprov.config import config

class MProvConnection:
    cache = None
    graph_name = None
    namespace = 'http://mprov.md2k.org'
    default_host = "localhost"
    QNAME_REGEX = re.compile('{([^}]*)}(.*)')

    """
    MProvConnection is a high-level API to the PennProvenance framework, with
    a streaming emphasis (i.e., tuples are stored with positions or timestamps,
    and derivations are recorded each time an action is invoked).
    """
    def __init__(self, user=None, password=None, host=None):
        # type: (str, str, str) -> None
        """
        Establish a connection to a PennProvenance server
        :param user: User ID
        :param password: Password for connection
        :param host: Host URL, or None to use localhost
        """
        if user is None:
            user = config.dbms.user

        if password is None:
            password = config.dbms.password

        if host is None:
            host = config.dbms.host

        #self.auth_conn = psycopg2.connect(host=host, database=config.dbms.auth_db, user=user, password=password)
        self.graph_conn = psycopg2.connect(host=host, database=config.dbms.graph_db, user=user, password=password)

        self.user_token = self.get_username()

        self._create_tables()
        self.graph_name = config.provenance.graph
        return

    def _create_tables(self):
        """
        Create the graph relations if need be
        """

        node_table = """
                     CREATE TABLE IF NOT EXISTS MProv_Node(_key VARCHAR(80) NOT NULL,
                                                           _resource VARCHAR(80) NOT NULL,
                                                           _created SERIAL,
                                                           label VARCHAR(80),
                                                           PRIMARY KEY(_resource, _key))
                     """

        node_props_table = """
                     CREATE TABLE IF NOT EXISTS MProv_NodeProp(_key VARCHAR(80) NOT NULL,
                                                           _resource VARCHAR(80) NOT NULL,
                                                           type VARCHAR(80),
                                                           label VARCHAR(80),
                                                           value VARCHAR,
                                                           code CHAR(1),
                                                           ivalue INTEGER,
                                                           lvalue BIGINT,
                                                           dvalue DOUBLE PRECISION,
                                                           fvalue REAL,
                                                           tvalue DATE,
                                                           tsvalue TIMESTAMP,
                                                           index BIGINT,
                                                           PRIMARY KEY(_resource, _key, label),
                                                           UNIQUE(_resource,_key,index),
                                                           FOREIGN KEY(_resource,_key) REFERENCES MProv_Node
                                                             ON DELETE CASCADE)
                           """

        edge_table = """
                     CREATE TABLE IF NOT EXISTS MProv_Edge(_key SERIAL,
                                                           _resource VARCHAR(80) NOT NULL,
                                                           _from VARCHAR(80) NOT NULL,
                                                           _to VARCHAR(80) NOT NULL,
                                                           label VARCHAR(80),
                                                           PRIMARY KEY(_resource, _key),
                                                           UNIQUE(_resource, _from, _to, label),
                                                           FOREIGN KEY(_resource, _from) REFERENCES MProv_Node
                                                            ON DELETE CASCADE,
                                                           FOREIGN KEY(_resource, _to) REFERENCES MProv_Node
                                                            ON DELETE CASCADE)
                     """

        edge_props_table = """
                     CREATE TABLE IF NOT EXISTS MProv_EdgeProp(_key INTEGER NOT NULL,
                                                           _resource VARCHAR(80) NOT NULL,
                                                           _created SERIAL,
                                                           type VARCHAR(80),
                                                           label VARCHAR(80),
                                                           value VARCHAR,
                                                           code CHAR(1),
                                                           ivalue INTEGER,
                                                           lvalue BIGINT,
                                                           dvalue DOUBLE PRECISION,
                                                           fvalue REAL,
                                                           tvalue DATE,
                                                           tsvalue TIMESTAMP,
                                                           index BIGINT,
                                                           PRIMARY KEY(_resource, _key, label),
                                                           UNIQUE(_resource,_key,index),
                                                           FOREIGN KEY(_resource,_key) REFERENCES MProv_Edge
                                                             ON DELETE CASCADE)
                           """

        schema_table = """
                     CREATE TABLE IF NOT EXISTS MProv_Schema(_key VARCHAR(80) NOT NULL,
                                                           _resource VARCHAR(80) NOT NULL,
                                                           name VARCHAR(80) NOT NULL,
                                                           value VARCHAR,
                                                           PRIMARY KEY(_resource, _key),
                                                           UNIQUE(_resource, name))
                       """

        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute(node_table)
                cursor.execute(node_props_table)
                cursor.execute(edge_table)
                cursor.execute(edge_props_table)
                cursor.execute(schema_table)
        return
    
    def create_or_reset_graph(self):
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("DELETE FROM MProv_Edge WHERE _resource = (%s)", (self.get_graph(),))
        try:
            self.store_agent(self.get_username())
        except psycopg2.errors.UniqueViolation:
            pass
        self.flush()

    def create_or_reuse_graph(self):
        try:
            self.store_agent(self.get_username())
        except psycopg2.errors.UniqueViolation:
            pass
        self.flush()

    def get_graph(self):
        """
        Within the storage system, get the name of the graph
        :return:
        """
        return self.graph_name

    def set_graph(self, name):
        """
        Set the name of the graph in the graph store
        :param name:
        :return:
        """
        self.graph_name = name

    def get_username(self):
        return config.provenance.user

    # Create a unique ID for an operator stream window
    @staticmethod
    def get_window_id(stream_operator, wid):
        # type: (str, Any) -> str
        return stream_operator + '_w.' + str(wid)

    # Create a unique ID for a stream operator result
    @staticmethod
    def get_result_id(stream, rid):
        # type: (str, Any) -> str
        return stream + '._r' + str(rid)

    # Create a unique ID for an entity
    @staticmethod
    def get_entity_id(stream, eid=None):
        # type: (str, Any) -> str
        if eid:
            return stream + '._e' + str(eid)
        else:
            return 'e_' + stream

    @staticmethod
    def get_stream_from_entity_id(eid):
        # type: (str) -> str
        if eid.startswith('{'):
            eid = MProvConnection.get_local_part(eid)

        if '._e' in eid:
            return eid[eid.index('._e')+3:]
        elif eid.startswith('e_'):
            return eid[2:]
        else:
            return eid

    # Create a unique ID for an entity
    @staticmethod
    def get_agent_id(user):
        # type: (str) -> str
        return 'u_' + user

    # Create a unique ID for an activity (a stream operator call)
    @staticmethod
    def get_activity_id(operator, aid):
        # type: (str, Any) -> str
        if aid:
            dk = hashlib.pbkdf2_hmac('sha256', (operator+aid).encode('utf-8'), b'prov', 20)

            stream = binascii.hexlify(dk).decode('utf-8')

            return stream#operator + '._a' + str(aid)
        else:
            dk = hashlib.pbkdf2_hmac('sha256', (operator).encode('utf-8'), b'prov', 20)

            stream = binascii.hexlify(dk).decode('utf-8')
            return stream#'a_' + operator

    def get_agent_token(self, agent_name):
        # type (str) -> str
        return  self.get_token_qname(self.get_agent_id(agent_name))

    def store_agent(self, agent_name):
        # type: (str) -> str
        agent_key = self._get_qname(self.user_token)
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Node(_key,_resource,label) VALUES(%s,%s,'AGENT') ON CONFLICT DO NOTHING", \
                    (agent_key,self.get_graph()))

                key = self._get_qname('agent_name')
                value = self.user_token

                cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,value,code,index) VALUES(%s,%s,%s,%s,'S',0) ON CONFLICT DO NOTHING", \
                    (agent_key, self.get_graph(), key, value))

        logging.debug('Storing AGENT %s' % str(self.user_token))

        return self.user_token

    def store_activity(self,
                       activity,
                       start,
                       end,
                       location= None):
        # type: (str, int, int, int) -> str
        """
        Create an entity node for an activity (a stream operator computation)

        :param activity: Name of the operation
        :param start: Start time
        :param end: End time
        :param location: Index position etc
        :return:
        """
        node_key = self.get_activity_id(activity,location)#self._get_qname(self.get_activity_id(activity,location))
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Node(_key,_resource,label) VALUES(%s,%s,'ACTIVITY') ON CONFLICT DO NOTHING", \
                    (node_key,self.get_graph()))

                key = self._get_qname('hash')
                value = activity

                cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,value,code,index) VALUES(%s,%s,%s,%s,'S',0) ON CONFLICT DO NOTHING", \
                    (node_key, self.get_graph(), key, value))

                key = self._get_qname('agent')
                value = self.get_username()

                cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,value,code,index) VALUES(%s,%s,%s,%s,'S',1) ON CONFLICT DO NOTHING", \
                    (node_key, self.get_graph(), key, value))

                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES((%s),(%s),(%s),'wasAssociatedWith') ON CONFLICT DO NOTHING", \
                    (self.get_graph(),node_key,self._get_qname(value)))

                key = self._get_qname('provDmStartTime')
                value = datetime.datetime.now()
                cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,tsvalue,code,index) VALUES(%s,%s,%s,%s,'S',2) ON CONFLICT DO NOTHING", \
                    (node_key, self.get_graph(), key, value))
                key = self._get_qname('provDmEndTime')
                value = datetime.datetime.now()
                cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,tsvalue,code,index) VALUES(%s,%s,%s,%s,'S',3) ON CONFLICT DO NOTHING", \
                    (node_key, self.get_graph(), key, value))


        token = node_key

        logging.debug('Storing ACTIVITY %s with ASSOCIATION %s', str(token), str(self.user_token))

        return token

    def _write_tuple(self, cursor, resource, node, tuple):
        if isinstance(tuple, BasicTuple):
            for i, k in enumerate(tuple.schema.fields):
                v = tuple[k]
                if isinstance(v, int):
                    cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,ivalue) VALUES(%s,%s,%s,%s) ON CONFLICT DO NOTHING", \
                        (node, resource, k, v))
                elif isinstance(v, float):
                    cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,fvalue) VALUES(%s,%s,%s,%s) ON CONFLICT DO NOTHING", \
                        (node, resource, k, v))
                elif isinstance(v, datetime.datetime):
                    cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,tsvalue) VALUES(%s,%s,%s,%s) ON CONFLICT DO NOTHING", \
                        (node, resource, k, v))
                else:
                    cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,value) VALUES(%s,%s,%s,%s) ON CONFLICT DO NOTHING", \
                        (node, resource, k, v))

                i = i + 1
        else:
            i = 0
            for k,v in tuple.items():
                if isinstance(v, int):
                    cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,ivalue) VALUES(%s,%s,%s,%s) ON CONFLICT DO NOTHING", \
                        (node, resource, k, v))
                elif isinstance(v, float):
                    cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,fvalue) VALUES(%s,%s,%s,%s) ON CONFLICT DO NOTHING", \
                        (node, resource, k, v))
                elif isinstance(v, datetime.datetime):
                    cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,tsvalue) VALUES(%s,%s,%s,%s) ON CONFLICT DO NOTHING", \
                        (node, resource, k, v))
                else:
                    cursor.execute("INSERT INTO MProv_NodeProp(_key,_resource,label,value) VALUES(%s,%s,%s,%s) ON CONFLICT DO NOTHING", \
                        (node, resource, k, v))

                i = i + 1

    def store_stream_tuple(self, stream_name, stream_index, input_tuple):
        # type: (str, int, BasicTuple) -> str

        """
        Create an entity node for a stream tuple

        :param stream_name: The name of the stream itself
        :param stream_index: The index position (count) or timestamp (if unique)
        :param input_tuple: The actual stream value
        :return: token for the new node
        """

        # The "token" for the tuple will be the node ID
        if isinstance(stream_index, int):
            token = self.get_token_qname(self.get_entity_id(stream_name, stream_index - 1))
        else:
            token = self.get_token_qname(self.get_entity_id(stream_name, stream_index))

        # We also embed the token within the tuple, in case the programmer queries the database
        # and wants to see it
        if isinstance(stream_index, int):
            input_tuple[self._get_qname("prov")] = self.get_entity_id(stream_name, stream_index - 1)
        else:
            input_tuple[self._get_qname("prov")] = self.get_entity_id(stream_name, stream_index)

        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Node(_key,_resource,label) VALUES(%s,%s,'ENTITY') ON CONFLICT DO NOTHING", \
                    (token,self.get_graph()))

                self._write_tuple(cursor, self.get_graph(), token, input_tuple)

        logging.debug('Storing ENTITY ' + str(token))

        return token

    def store_code(self, code):
        # type: (str) -> str

        """
        Store a code definition as an entity

        :param code: Source code definition
        :return: String ID (local part of QName) for the node
        """

        dk = hashlib.pbkdf2_hmac('sha256', code.encode('utf-8'), b'mprov', 10000)

        stream = binascii.hexlify(dk).decode('utf-8')

        # The "token" for the tuple will be the node ID
        token = self.get_entity_id(stream)

        # Now we'll create a tuple within the provenance node, to capture the data
        data = {self._get_qname("prov"): token, 'code': code, 'type': 'python3'}

        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Node(_key,_resource,label) VALUES(%s,%s,'ENTITY') ON CONFLICT DO NOTHING", \
                    (token,self.get_graph()))

                self._write_tuple(cursor, self.get_graph(), token, data)

        logging.debug('Storing ENTITY ' + str(token))

        return token

    def store_annotations(self,
                          node_token,
                          annotation_dict):
        # type: (str, dict) -> List[str]
        """
        Associate a map of annotations with a node ID

        :param node_token:
        :param annotation_dict:
        :return:
        """

        ann_tokens = []

        for k in annotation_dict.keys():
            ann_token = self.get_token_qname(MProvConnection.get_local_part(node_token) + "." + k)
            ann_tokens.append(ann_token)

            # The key/value pair will itself be an entity node
            attributes = {k: annotation_dict[k]}
            self._write_annot(ann_token, node_token, attributes)

        return ann_tokens

    def _write_annot(self, ann_token, node_token, attributes):
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Node(_key,_resource,label) VALUES(%s,%s,'ENTITY') ON CONFLICT DO NOTHING", \
                    (ann_token,self.get_graph()))
                # Then we add a relationship edge (of type ANNOTATED)
                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'_annotated') ON CONFLICT DO NOTHING", \
                    (self.get_graph(),ann_token,node_token))
                logging.debug('Wrote ANNOT edge %s' % ann_token)

                # Write the annotation tuple
                self._write_tuple(cursor, self.get_graph(), ann_token, attributes)


    def store_annotation(self,
                         stream_name,
                         stream_index,
                         annotation_name,
                         annotation_value):
        # type: (str, int, str, Any) -> str
        """
        Create a node for an annotation to an entity / tuple

        :param stream_name: The name of the stream itself
        :param stream_index: The index position (count) or timestamp (if unique) of the
                stream element we are annotating
        :param annotation_name: The name of the annotation
        :param annotation_value: The value of the annotation
        :return:
        """

        # The "token" for the tuple will be the node ID
        node_token = self.get_token_qname(self.get_entity_id(stream_name, stream_index - 1))

        ann_token = self.get_token_qname(self.get_entity_id(stream_name + '.' + annotation_name, stream_index - 1))

        # The key/value pair will itself be an entity node
        attributes = {annotation_name: annotation_value}

        self._write_annot(ann_token, node_token, attributes)

        return ann_token

    def store_window_and_inputs(self,
                                output_stream_name,
                                output_stream_index,
                                input_tokens_list
                                ):
        # type: (str, int, list) -> str
        """
        Store a mapping between an operator window, from
        which a stream is to be derived, and the input
        nodes

        :param output_stream_name:
        :param output_stream_index:
        :param input_tokens_list:
        :return:
        """

        # The "token" for the tuple will be the node ID
        if isinstance(output_stream_index, int):
            window_token = self.get_token_qname(self.get_window_id(output_stream_name, output_stream_index - 1))
        else:
            window_token = self.get_token_qname(self.get_window_id(output_stream_name, output_stream_index))

        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Node(_key,_resource,label) VALUES(%s,%s,'COLLECTION') ON CONFLICT DO NOTHING", \
                    (window_token,self.get_graph()))

                logging.debug('Storing COLLECTION %s' % str(window_token))

                for token in input_tokens_list:
                    # Add a relationship edge (of type ANNOTATED)
                    # from window to its inputs
                    token_qname = self.get_token_qname(token)

                    cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'hadMember') ON CONFLICT DO NOTHING", \
                        (self.get_graph(), window_token, token_qname))

        return window_token

    def store_derived_result(self,
                             output_stream_name,
                             output_stream_index,
                             output_tuple,
                             input_token,
                             activity,
                             start,
                             end
                             ):
        # type: (str, int, BasicTuple, list, str, int, int) -> str
        """
        When we have a windowed computation, this creates a complex derivation subgraph
        in one operation.

        :param output_stream_name: The name of the stream our operator produces
        :param output_stream_index: The position of the outgoing tuple in the stream
        :param output_tuple: The tuple itself
        :param input_token: IDs of the inputs to the computation
        :param activity: The computation name
        :param start: Start time
        :param end: End time
        :return:
        """
        result_token = self.store_stream_tuple(output_stream_name, output_stream_index, output_tuple)

        activity_token = self.store_activity(activity, start, end, output_stream_index)

        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'wasDerivedFrom') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), result_token, input_token))

                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'used') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), activity_token, input_token))

                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'wasGeneratedBy') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), result_token, activity_token))

        logging.debug('Storing DERIVATION %s', result_token)
        return result_token

    def create_collection(self,
                          collection_name, collection_version=None,
                          prior_token=None):
        # type: (str, int, str) -> str
        """
        We can create a collection to represent a sub-sequence, a sub-stream, or a subset of items

        :param collection_name:
        :param collection_version:
        :param prior_token:
        :return:
        """
        token = self.get_token_qname(self.get_entity_id(collection_name, collection_version))

        # Create a node for the collection
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Node(_key,_resource,label) VALUES(%s,%s,'COLLECTION') ON CONFLICT DO NOTHING", \
                    (token,self.get_graph()))

        if prior_token:
            with self.graph_conn as conn:
                with conn.cursor() as cursor:
                    cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'wasDerivedFrom') ON CONFLICT DO NOTHING", \
                        (self.get_graph(), token, prior_token))

        return token

    def add_to_collection(self, tuple_token, collection_token):
        # type (str, str) -> str
        """
        Associate a tuple with a collection (using the tokens for each)

        :param tuple_token:
        :param collection_token:
        :return:
        """
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'hadMember') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), collection_token, tuple_token))

    def store_windowed_result(self,
                              output_stream_name,
                              output_stream_index,
                              output_tuple,
                              input_tokens_list,
                              activity,
                              start,
                              end
                              ):
        # type: (str, int, BasicTuple, list, str, int, int) -> str
        """
        When we have a windowed computation, this creates a complex derivation subgraph
        in one operation.

        :param output_stream_name: The name of the stream our operator produces
        :param output_stream_index: The position of the outgoing tuple in the stream
        :param output_tuple: The tuple itself
        :param input_tokens_list: IDs of the inputs to the computation
        :param activity: The computation name
        :param start: Start time
        :param end: End time
        :return:
        """
        result_token = self.store_stream_tuple(output_stream_name, output_stream_index, output_tuple)
        window_token = self.store_window_and_inputs(output_stream_name, output_stream_index, input_tokens_list)

        activity_token = self.store_activity(activity, start, end, output_stream_index)

        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'wasDerivedFrom') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), result_token, window_token))
                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'used') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), activity_token, window_token))

                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'wasGeneratedBy') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), result_token, activity_token))

        return result_token

    def store_derived_from(self, derived_node, source_node):
        # types: (str, str) -> str
        """
        Stores a derivation of an output from an input
        :param derived_node:
        :param source_node:
        :return:
        """
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'wasDerivedFrom') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), derived_node, source_node))

    def store_used(self, activity_node, input_node):
        # types: (str, str) -> str
        """
        Stores a usage relationship between an activity and an input
        :param activity_node:
        :param input_node:
        :return:
        """
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'used') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), activity_node, input_node))

    def store_generated_by(self, output_node, activity_node):
        # types: (str, str) -> str
        """
        Stores an edge indicating an output was generated by an activity
        :param self:
        :param output_node:
        :param activity_node:
        :return:
        """
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("INSERT INTO MProv_Edge(_resource,_from,_to,label) VALUES(%s,%s,%s,'wasGeneratedBy') ON CONFLICT DO NOTHING", \
                    (self.get_graph(), output_node, activity_node))

    def _get_qname(self, local_part):
        # types: (str) -> str
        """
        Returns a qualified name from a string to be used as the local part
        :param local_part:
        :return:
        """
        return "{" + self.namespace + "}" + local_part

    def get_token_qname(self, token):
        # types: (str) -> str
        """
        Returns a qualified name created by hashing the given token. Used for cases where the length of
        the resulting name needs to be of limited length, for example when creating a prov token.
        :param token:
        :return:
        """
        if len(token) > 40:
            thash = hashlib.sha1(token.encode('utf-8'))
            return self._get_qname(thash.hexdigest())
        else:
            return self._get_qname(token)

    def get_node(self, entity_id):
        # type: (str) -> List[Dict]
        """
        Returns the tuple data associated with a node ID (generally an entity node)
        :param entity_id:
        :return:
        """
        result = self.get_provenance_data(self.get_graph(), (entity_id))

        return result

    def get_source_entities(self, entity_id):
        # type: (str) -> List[str]
        """
        Given an entity node, what was it derived from?
        :param entity_id:
        :return:
        :param entity_id:
        :return:
        """
        results = self.get_connected_from(self.get_graph(), (entity_id), 'wasDerivedFrom')
        # results = [self.parse_qname(tok.token_value) for tok in results.tokens]
        return results

    def get_derived_entities(self, entity_id):
        # type: (str) -> List[str]
        results = self.get_connected_to(self.get_graph(), (entity_id), 'wasDerivedFrom')

        return results

    def get_parent_entities(self, entity_id):
        # type: (str) -> List[str]
        """
        Find a parent collection (ie a node where this node is a member)
        :param entity_id:
        :return:
        """
        results = self.get_connected_to(self.get_graph(), (entity_id), 'hadMember')
        # results = [self.parse_qname(tok.token_value) for tok in results.tokens]
        return results

    def get_child_entities(self, entity_id):
        # type: (str) -> List[str]
        """
        Find any child (member) nodes
        :param entity_id:
        :return:
        """
        results = self.get_connected_from(self.get_graph(), (entity_id), 'hadMember')
        # results = [self.parse_qname(tok.token_value) for tok in results.tokens]
        return results

    def get_creating_activities(self, entity_id):
        # type: (str) -> List[str]
        """
        Get the list of activities that from which this output was generated
        :param entity_id:
        :return:
        """
        results = self.get_connected_from(self.get_graph(), (entity_id), 'wasGeneratedBy')

        return results

    def get_activity_outputs(self, activity_id):
        # type: (str) -> List[str]
        """
        Get the list of entities this activity generated
        :param activity_id:
        :return:
        """
        results = self.get_connected_to(self.get_graph(), (activity_id), 'wasGeneratedBy')
        return results

    def get_activity_inputs(self, activity_id):
        # type: (str) -> List[str]
        """
        Given an activity, what inputs did it use?
        :param activity_id:
        :return:
        """
        results = self.get_connected_from(self.get_graph(), (activity_id), 'used')
        # results = [self.parse_qname(tok.token_value) for tok in results.tokens]
        return results

    def get_annotations(self, node_id):
        # type: (str) -> List[dict]
        """
        Given a node, find its annotations and return as a list of typed key/values
        :param node_id:
        :return:
        """
        results = self.get_connected_to(self.get_graph(), (node_id), '_annotated')

        results = [self.get_node(eid) for eid in results]

        return results

    def get_stream_inputs(self, stream_name):
        inputs = []

        stream_node = self.get_token_qname(self.get_entity_id(stream_name))

        for node in self.get_source_entities(stream_node):
            inputs.append(self.get_stream_from_entity_id(MProvConnection.get_local_part(node)))

        return inputs

    def get_stream_producers(self, stream_name):
        producers = []
        stream_node = self.get_token_qname(self.get_entity_id(stream_name))

        for node in self.get_creating_activities(stream_node):
            code = self.get_node(node)

            if len(code) > 0:
                code_id  = code[0][0]#[self._get_qname('hash')]
                #print(code_id)
                code = self.get_node(code_id)#self.get_token_qname(code_id))
                producers.append(self._get_qname(code[0]['code']))

        return producers

    @classmethod
    def get_local_part(cls, token_value):
        # types (str) -> str
        """
        Returns a QualifiedName by parsing the given string as a namespace and local part
        :param token_value: a string of the form '{' + namespace + '}' + local_part
        :return: the corresponding local portion of the name
        """
        match = cls.QNAME_REGEX.match(token_value)
        if not match:
            raise ValueError('cannot parse as QualfiedName:', token_value)
        return match.group(2)

    def get_provenance_data(self, resource, token):
        # type: (str, str) -> List[Dict]
        ret = []
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                cursor.execute("SELECT index,code,value,ivalue,lvalue,fvalue,dvalue,tvalue,tsvalue,label FROM MProv_NodeProp WHERE _resource = (%s) AND _key = (%s)", (resource,token))

                results = cursor.fetchall()
                ret = {}#[None for i in range(0,len(results))]
                for res in results:
                    #print(res)
                    inx = res[0]
                    if inx is None:
                        inx = res[9]
                    if res[1] is None or res[1] == 'S':
                        ret[inx] = res[2]
                    elif res[1] == 'I':
                        ret[inx] = res[3]
                    elif res[1] == 'L':
                        ret[inx] = res[4]
                    elif res[1] == 'F':
                        ret[inx] = res[5]
                    elif res[1] == 'D':
                        ret[inx] = res[6]
                    elif res[1] == 'T':
                        ret[inx] = res[7]
                    elif res[1] == 't':
                        ret[inx] = res[8]
                    else:
                        raise RuntimeError('Unknown code ' + res[1])


                #print ("Results:", token, cursor.fetchall())

        return [ret]

    def get_connected_to(self, resource, token, label1):
        # type: (str, str, str) -> List[pennprov.ProvTokenSetModel]
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                if label1 is None:
                    cursor.execute("SELECT _from FROM MProv_Edge WHERE _resource = (%s) AND _to = (%s)", (resource,token))
                else:
                    cursor.execute("SELECT _from FROM MProv_Edge WHERE _resource = (%s) AND _to = (%s) AND label = (%s)",
                                   (self.get_graph(), token, label1))
                return [x[0] for x in cursor.fetchall()]
        return []

    def get_connected_from(self, resource, token, label1):
        # type: (str, str, str) -> List[pennprov.ProvTokenSetModel]
        with self.graph_conn as conn:
            with conn.cursor() as cursor:
                if label1 is None:
                    cursor.execute("SELECT _to FROM MProv_Edge WHERE _resource = (%s) AND _from = (%s)", (resource,token))
                else:
                    cursor.execute("SELECT _to FROM MProv_Edge WHERE _resource = (%s) AND _from = (%s) AND label = (%s)",
                                   (self.get_graph(), token, label1))
                return [x[0] for x in cursor.fetchall()]
        return []

    def flush(self):
        return

    def close(self):
        return

    def __del__(self):
        self.close()
    



import pytest
import logging

import pennprov.connection.mprov as mprov

class TestMProv:
    conn = None

    def test_create(self):
        self.conn = mprov.MProvConnection()
        assert self.conn is not None

        self.conn.create_or_reset_graph()
        self.conn.create_or_reuse_graph()

        self.conn.store_stream_tuple('first_stream', 0, {'start': 1, 'end': 2})
        self.conn.store_annotation('first_stream', 0, 'ann1', 30)

        self.conn.store_activity('area_circle', 0, 1, 0)
        self.conn.store_code('import pytest\nimport logging\n')






import logging
import sys
import traceback
from datetime import timezone, datetime
from functools import wraps

from pennprov.connection.mprov import MProvConnection
from pennprov.connection.mprov_connection_cache import MProvConnectionCache
from pennprov.metadata.stream_metadata import BasicSchema, BasicTuple
import pandas as pd
import inspect

blank = BasicSchema({})
blank_tuple = BasicTuple(blank)

get_entity_id = MProvConnection.get_entity_id

stored = {}
class MProvAgg:
    def __init__(self, in_stream_name,out_stream_name,in_stream_key=['index'],out_stream_key=['index'],collection=None,connection_key=None):
        """
        MProvAgg: decorator for an aggregation operation over windows.  Creates a provenance
        node for each output stream element and attaches it to the in_stream_keys for the inputs.

        :param in_stream_name: The name of the input stream schema
        :param out_stream_name: The name of the output stream schema
        :param in_stream_key: The key fields in the input stream
        :param out_stream_key: The key fields of the output stream
        :param collection: The name of the output stream collection, to which this result should be appended
        :param connection_key: Unique ID for reusing cached connection to MProv server
        :return:
        """
        self.in_stream_name = in_stream_name
        self.out_stream_name = out_stream_name
        self.in_stream_key = in_stream_key
        self.out_stream_key = out_stream_key
        self.collection = collection
        self.connection_key = connection_key or MProvConnectionCache.Key()

        # For Spark. Flush anything created on the driver node prior to computation.
        mprov_conn = MProvConnectionCache.get_connection(self.connection_key);
        if mprov_conn:
            mprov_conn.flush()

    def md_key(self, stream, key_list):
        if isinstance(stream, tuple) and len(stream) == 1:
            stream = stream[0]
        # If the stream is a dataframe, return the subset matching the key
        if isinstance(stream, pd.DataFrame):
            if not self.in_stream_key:
                return stream
            else:
                return str(stream.reset_index()[key_list])
        else:
            return repr(stream)

    def rel_keys(self, stream, key_list):
        if isinstance(stream, tuple) and len(stream) == 1:
            stream = stream[0]
        # If the stream is a dataframe, return the subset matching the key
        if isinstance(stream, pd.DataFrame):
            if not self.in_stream_key:
                return stream
            else:
                return stream.reset_index()[key_list].to_dict('records')
        else:
            return repr(stream)

    @staticmethod
    def _write_collection_relationships(mprov_conn, sig, derived, source):
        mprov_conn.store_derived_from(derived, source)
        activity_token = mprov_conn.store_activity(sig, None, None, None)

        mprov_conn.store_used(activity_token, source)
        mprov_conn.store_generated_by(derived, activity_token)

    def __call__(self, func):
        global stored
        # Need this to happen here (decoration-time and not call-time)
        # while the source file is still available to Spark.
        if func.__name__ not in stored:
            d = (inspect.getsource(func))
            logging.debug('Read source of %s as %s', func.__name__, d)

        @wraps(func)
        def wrapper(arg):
            #args_repr = [md_key(a, in_stream_key) for a in args]
            #kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
            #signature = ", ".join(args_repr + kwargs_repr)
            #print(f"Calling {func.__name__}({signature})")

            val = func(arg)#*args, **kwargs)

            mprov_conn = MProvConnectionCache.get_connection(self.connection_key)

            global stored

            stored[func.__name__] = mprov_conn.store_code(d)
            logging.debug('Storing %s as %s:%s', func.__name__, stored[func.__name__], d)
            sig = stored[func.__name__]




            # Create a window with the appropriate keys
            window_ids = []
            #for t in self.rel_keys(arg, self.in_stream_key):
            #    window_ids.append(get_entity_id(self.in_stream_name, [t[k] for k in self.in_stream_key]))

            if isinstance(self.out_stream_key,list):
                if len(self.out_stream_key) == 1:
                    out_keys = val.reset_index()[self.out_stream_key[0]].to_list()
                else:
                    out_keys = []
                    # print(val.reset_index()[out_stream_key])
                    for tup in val.reset_index()[self.out_stream_key].iterrows():
                        out_keys.append(tup[1].to_list())
            else:
                out_keys = val.reset_index()[self.out_stream_key].to_list()

            ts = datetime.now(timezone.utc)
            if mprov_conn:
                try:
                    for t in self.rel_keys(arg, self.in_stream_key):
                        #print('Input: %s' %self.in_stream_name + str([t[k] for k in self.in_stream_key]))
                        window_ids.append(get_entity_id(self.in_stream_name, [t[k] for k in self.in_stream_key]))

                        # This ensures that the input tuples actually exist.  It needs to merge
                        # if the tuple is already there
                        tup = mprov_conn.store_stream_tuple(self.in_stream_name,
                                                            [t[k] for k in self.in_stream_key], blank_tuple)

                        # Ensure that the input tuples are also linked to the appropriate stream
                        if self.in_stream_name not in stored:
                            stream_part = mprov_conn.create_collection(self.in_stream_name)
                            stored[self.in_stream_name] = MProvConnection.get_local_part(stream_part)
                        else:
                            stream_part = mprov_conn.get_token_qname(stored[self.in_stream_name])
                        mprov_conn.add_to_collection(tup, stream_part)
                except:
                    print('Error')
                    print(sys.exc_info()[0])
                    traceback.print_exc()
                    pass
                #print (window_ids)
                result = mprov_conn.store_windowed_result(self.out_stream_name, 'w'+str(out_keys), blank_tuple,
                                                 window_ids, sig, ts, ts)
                if self.out_stream_name not in stored:
                    stream_part = mprov_conn.create_collection(self.out_stream_name)
                    stored[self.out_stream_name] = MProvConnection.get_local_part(stream_part)

                    # Add derivation to source input
                    self._write_collection_relationships(mprov_conn, sig, stream_part, mprov_conn.get_token_qname(stored[self.in_stream_name]))
                else:
                    stream_part = mprov_conn.get_token_qname(stored[self.out_stream_name])
                mprov_conn.add_to_collection(result, stream_part)

                if self.collection:
                    mprov_conn.add_to_collection(result, self.collection)
            else:
                logging.warning("Output: %s.%s <(%s)- %s", self.out_stream_name, 'w'+str(out_keys), sig, str(window_ids))

            return val
        return wrapper

#if __name__ == '__main__':




import pennprov
from pennprov.connection.mprov import MProvConnection
import hashlib, binascii
import sqlparse
import pyspark

class SqlProvenance:
    tab_to_schema = {}
    mprov = None

    def __init__(self, mprov_conn):
        # type: (MProvConnection) -> None
        self.tab_to_schema = {}
        self.mprov = mprov_conn

    def query_to_entity(self, query):
        # type: (str) -> pennprov.QualifiedName
        """
        Takes an SQL query and gets a QName to an entity
        :param query:
        :return:
        """
        return self.mprov.store_code(query)

    def create_or_replace_temp_view(self, py_name, sql_name):
        # type: (str, str) -> None
        """
        Replacement for spark.createOrReplaceTempView(), logs
        naming and schema info

        :param py_name: Python variable
        :param sql_name: Table name for SQL
        :return:
        """
        self.tab_to_schema[sql_name] = py_name.schema
        py_name.createOrReplaceTempView(sql_name)

        print(self.tab_to_schema[sql_name])




# coding: utf-8

# flake8: noqa

"""
    Habitat repository and authorization API

    Habitat API  # noqa: E501

    OpenAPI spec version: V1.1.2
    Contact: zives@seas.upenn.edu
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from __future__ import absolute_import

# import apis into sdk package
from pennprov.api.authentication_api import AuthenticationApi
from pennprov.api.permission_api import PermissionApi

# import ApiClient
from pennprov.api_client import ApiClient
from pennprov.configuration import Configuration
from pennprov.models.user_credentials import UserCredentials
from pennprov.models.user_info import UserInfo
from pennprov.models.web_token import WebToken



'''
 Copyright 2020 Trustees of the University of Pennsylvania

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''
import os
import logging
from pennprov.connection.mprov import MProvConnection
from urllib3.exceptions import NewConnectionError
from pennprov.config import config

logger = logging.getLogger(__name__)


class MProvConnectionCache:
    """
    A simple in-memory cache for MProvConnections
    where the connections are stored in a class variable dict.
    """
    class Key:
        """
        A key for the connection cache. Instances should not be modified after instantiation.
        """

        def __init__(self, graph=MProvConnection.graph_name, user=None, password=None, host=None):
            # type: (str, str, str, str) -> MProvConnectionCache.Key
            """
            :param graph:    The name of the graph with which this connection will be interacting.
                             Defaults to MProvConnection.graph_name
            :param user:     The username to use for the connection.
                             Defaults to the value of the environment variable MPROV_USER
            :param password: The password to use for the connection.
                             Defaults to the value of the environment variable MPROV_PASSWORD
            :param host:     The host to use for the connection, e.g., 'http://localhost:1234'
                             Defaults to the value of the environment variable MPROV_HOST if set,
                             otherwise to MProvConnection.default_host
            :raises:         ValueError if user is None and MPROV_USER is not set
                             or if password is None and MPROV_PASSWORD is not set
            """
            if graph is not None:
                self.graph = graph
            else:
                self.graph = config.provenance.graph
            self.user = user or os.environ.get('MPROV_USER')
            if self.user is None:
                self.user = config.dbms.user
            self.password = password or os.environ.get('MPROV_PASSWORD')
            if self.password is None:
                self.password = config.dbms.password
            self.host = host or os.environ.get(
                'MPROV_HOST', MProvConnection.default_host)
            if self.host is None:
                self.host = config.dbms.host
            if not self.user:
                raise ValueError(
                    'either supply user or set environment variable MPROV_USER')
            if not self.password:
                raise ValueError(
                    'either supply password or set environment variable MPROV_PASSWORD')

        def __hash__(self):
            return hash((self.graph, self.user, self.password, self.host))

        def __eq__(self, other):
            return (self.graph, self.user, self.password, self.host) == (other.graph, other.user, other.password, other.host)

        def __ne__(self, other):
            return not self.__eq__(other)

    connections = {}

    @classmethod
    def get_connection(cls, connection_key):
        # type: (MProvConnectionCache.Key) -> MProvConnection
        """
        Returns a cached MProvConnection for the given key if one has been previously cached.
        Otherwise, attempts to create, cache, and return a newly created connection.
        If the connection cannot be created, None is returned and the error is logged at the DEBUG level.

        :param connection_key: an MProvConnectionCache.Key
        :return: the connection associated with connection_key or a newly created connection
        """
        connection = cls.connections.get(connection_key, None)
        if connection:
            logger.debug(
                'MProvConnectionCache: process %s retrieved cached connection %s', os.getpid(), id(connection))
        else:
            try:
                connection = MProvConnection(
                    connection_key.user, connection_key.password, connection_key.host)
                connection.set_graph(connection_key.graph)

                logger.debug(
                    'MProvConnectionCache: process %s created new connection %s', os.getpid(), id(connection))
            except Exception as e:
                logger.debug(e, exc_info=True)
                connection = None

            if connection:
                cls.connections[connection_key] = connection
        return connection



"""
 Copyright 2019 Trustees of the University of Pennsylvania
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""

from typing import Any

class BasicSchema:
    count = 0
    fields = []
    types = []

    def __init__(self, *args):
        if len(args) == 1:
            self.init1(args[0])
        elif len(args) == 2:
            self.init2(args[0], args[1])
        elif len(args) == 3:
            self.init3(args[0], args[1], args[2])
        else:
            raise TypeError('Illegal inputs')

    def init1(self, name):
        # type: (str) -> None
        self.name = name
        self.fields = []
        self.types = []

    def init2(self, name, fields_types):
        # type: (str, dict) -> None
        """ Initialize with a name and a dictionary of attributes to types """
        self.name = name
        self.fields = []
        self.types = []
        for k in fields_types.keys():
            self.fields.append(k)
            self.types.append(fields_types[k])

    def init3(self, name, fields, types):
        # type: (str, list, list) -> None
        """ Initialize with a name and a list of field-type pairs """
        self.name = name
        self.fields = fields
        self.types = types

    def add_field(self, k, v):
        # type: (str, Any) -> None
        self.fields.append(k)
        self.types.append(v)

    def get_field(self, k):
        # type: (str) -> Any
        i = self.fields.index(k)
        if i:
            return self.types[i]
        else:
            return None

    def get_name(self):
        return self.name

    def create_tuple(self, *args):
        if len(args) == 0:
            return self.create_tuple_blank()
        elif isinstance(args[0], list):
            return self.create_tuple_list(args[0])
        else:
            return self.create_tuple_dict(args[0])

    def create_tuple_blank(self):
        return BasicTuple(self)

    def create_tuple_dict(self, content):
        # type: (dict) -> BasicTuple
        return BasicTuple(self, content)

    def create_tuple_list(self, content):
        # type (list) -> BasicTuple
        dict2 = {}
        for i, k in enumerate(self.fields):
            dict2[k] = content[i]

        return BasicTuple(self, dict2)

    def __str__(self):
        ret = self.name + '('
        for i, f in enumerate(self.fields):
            if i > 0:
                ret = ret + ','
            ret = ret + f + ':' + self.types[i]

        return ret + ")"

class BasicTuple:
    schema = None
    data = {}

    def __init__(self, *args):
        if len(args) == 1:
            self.init1(args[0])
        elif len(args) == 2:
            self.init2(args[0], args[1])
        else:
            raise TypeError('Illegal inputs')

    def init1(self, schema):
        # type: (BasicSchema) -> None
        self.schema = schema
        for i, name in enumerate(schema.fields):
            self.data[name] = None

    def init2(self, schema, values):
        # type: (BasicSchema, dict) -> None
        self.schema = schema
        self.data = values

    def __str__(self):
        ret = self.schema.get_name() + '('
        for i, f in enumerate(self.schema.fields):
            if i > 0:
                ret = ret + ','
            if f is None:
                raise TypeError("Can't have a null key")
            elif self.data is None or self.data[f] is None:
                ret = ret + f
            else:
                ret = ret + f + ':' + str(self.data[f])

        return ret + ")"

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.data[self.schema.fields[item]]
        else:
            return self.data[item]

    def __setitem__(self, key, value):
        if key in self.schema.fields:
            self.data[key] = value
        elif isinstance(key, int):
            self.data[self.schema.fields[key]] = value



# coding: utf-8

# flake8: noqa
"""
    Habitat repository and authorization API

    Habitat API  # noqa: E501

    OpenAPI spec version: V1.1.2
    Contact: zives@seas.upenn.edu
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from __future__ import absolute_import

# import models into model package
from pennprov.models.response_error import ResponseError
from pennprov.models.user_credentials import UserCredentials
from pennprov.models.user_info import UserInfo
from pennprov.models.web_token import WebToken
